#include "search.h"
#include "../simulatorNetworkInterface/simulatorNetworkInterface.h"
#include "../storage/storage.h"
#include "../statistics/statistics.h"
#include "../datalog/datalog.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>

int erraticPrune = 0;
int normalPrune = 0;
int sameDelete = 0;
int otherDelete = 0;

/**
 * Mock function to return time
 * @return Current mocked time
 */
uint16_t getCurrentTime() {
#ifdef TESTING
    return 10;
#endif
    return simulatorRequestTime();
}

double der3(uint32_t cell) {
    struct dataPoint *last = storageGetLatestDataPoint(cell);
    double der3 = 0;
    uint8_t nmbOfDataPoints = storageGetNmbOfDataPoints(cell);
    if (nmbOfDataPoints > 2) {
        // Der3 can be generated
        dataPoint *p = storageGetDatapointNmbX(cell, nmbOfDataPoints - (uint8_t) 3);
        der3 = (double) (last->data.meas - p->data.meas) / (last->data.time - p->data.time);
    }
    return der3;
}

double totalDerivative(uint32_t cell) {
    struct dataPoint *first = storageGetFirstDataPoint(cell);
    struct dataPoint *last = storageGetLatestDataPoint(cell);
    if (first == last) {
        return 0;
    }
    double totalDerivative =
            (double) (last->data.meas - first->data.meas) / (last->data.time - first->data.time);
    return totalDerivative;
}

int16_t absoluteChange(uint32_t cell) {
    struct dataPoint *first = storageGetFirstDataPoint(cell);
    struct dataPoint *last = storageGetLatestDataPoint(cell);
    int16_t absoluteChange = (uint16_t) (first->data.meas - last->data.meas);
    if (absoluteChange > 1000) {
        absoluteChange = 1000;
    } else if (absoluteChange < -1000) {
        absoluteChange = -1000;
    }
    return absoluteChange;
}

int pruneSearch(){
    uint32_t existingCells[100];
    int numberOfExistingCells = storageGetListOfCells(existingCells, 100);
    int prunedErraticCell = 0;
    for(int i = 0; i < numberOfExistingCells; i++){
        if(pruneCellErratic(existingCells[i]) == 1){
            prunedErraticCell = existingCells[i];
        }
        pruneCellSimple(existingCells[i]);
    }
    return prunedErraticCell;
}

void considerNewDatapoint(dataPoint *p, uint32_t cell) {
    if (storageGetNmbEmptyBlocks() > 0) {
        // Free space, add
        printf("Free for cell %d \t remaining blocks : %d\n", cell, storageGetNmbEmptyBlocks());
        assert(storageAddDatapoint(p, cell) == 1);
        return;
    }

    // Not free space, prune. If pruned same cell, continue
    if (pruneSearch() == cell) return;

    if (storageGetNmbEmptyBlocks() > 0) {
        // Free space, add
        printf("Free after prune for %d\n", cell);
        assert(storageAddDatapoint(p, cell) == 1);
        return;
    }

    // No free space after pruning. How many cells do we have stored?
    uint8_t nmbOfCells = storageGetNmbOfCells();
    dataPoint *oldPoint;
    double c = datapointValue(p, cell, 1);
    if (nmbOfCells <= PREFERED_NMB_OF_CELLS) {
        // There must be some cells with more than MAX_NMB_OF_POINTS_FOR_CELL points
        uint8_t nmbOfPoints = storageGetNmbOfDataPoints(cell);
        if (nmbOfPoints >= MAX_NMB_OF_POINTS_FOR_CELL) {
            // This is the cell with more datapoints or equal, delete lowest valued
            double l = storageFindLowestValuedDatapoint(cell, &oldPoint);
            if (c > l) {
                // New datapoint has greater value
                assert(storageDeleteDatapoint(cell, oldPoint->data.time) == 1);
                assert(storageAddDatapoint(p, cell) == 1);
                printf("Same cell delete for %d\n", cell);
            } else {
                // New datapoint does not have greater value, remove the last point and insert new point
                assert(storageDeleteDatapoint(cell, (uint16_t) storageGetLatestDataPoint(cell)->data.time) == 1);
                assert(storageAddDatapoint(p, cell) == 1);
                printf("Added last point same cell %d\n", cell);
            }
            sameDelete++;
            return;
        }

        // Need to find the cell with more than MAX_NMB_OF_POINTS_FOR_CELL points. Delete it and insert new point
        uint32_t cellWithManyPoints;
        uint8_t nmbOfManyPoints;
        storageFindCellWithMostPoints(&cellWithManyPoints, &nmbOfManyPoints);
        storageFindLowestValuedDatapoint(cellWithManyPoints, &oldPoint);
        printf("Deleted point t = %d for cell %d  for new point %d\n", oldPoint->data.time, cellWithManyPoints, cell);
        otherDelete++;
        assert(storageDeleteDatapoint(cellWithManyPoints, oldPoint->data.time) == 1);
        assert(storageAddDatapoint(p, cell) == 1);
    } else {
#ifdef MODE2
        // There exist more than allowed cells, delete lowest valued point for lowest valued cell
        // This is an edge case. It should rarely happen, as there would be many failing cells up to this point.
        uint32_t lowValuedCell;
        double value;
        storageFindLowestValuedCell(&lowValuedCell, &value);
        dataPoint *lowValuedPoint;
        value = storageFindLowestValuedDatapoint(lowValuedCell, &lowValuedPoint);
        if(cell == lowValuedCell){
            // Already on on the lowest value cell, do not add
            printf("Cell %d is already worst, skip\n", cell);
            return;
        }
        if (value == INFINITY) {
            // There only exists this one point, delete it
            assert(storageDeleteCell(lowValuedCell) == 1);
            printf("More than allowed cells. Deleted cell %d for new point for cell %d\n", lowValuedCell, cell);
        } else {
            int r = storageDeleteDatapoint(lowValuedCell, lowValuedPoint->data.time);
            printf("More than allowed cells. Deleted point t = %d for cell %d for new point %d\n",
                   lowValuedPoint->data.time, lowValuedCell, cell);
            assert(r == 1);
        }
        assert(storageAddDatapoint(p, cell) == 1);
#endif
    }
}

void watchRoutine(uint32_t cell, uint16_t icell) {
    dataPoint p;
    p.data.meas = icell;
    p.data.time = getCurrentTime();
    if (storageDoesDatapointsExist(cell)) {
        struct dataPoint *latest = storageGetLatestDataPoint(cell);
        if ((p.data.time - latest->data.time) > SEARCH_MIN_DELTA_T) {
            // New datapoint found as more than SEARCH_MIN_DELTA_T has passed. Consider it
            considerNewDatapoint(&p, cell);
        } else {
            // Data point must be older.
            return;
        }
    } else {
        considerNewDatapoint(&p, cell);
    }
}

void searchPartOfFlash(uint16_t startWord, uint16_t endWord) {
    uint32_t word[32];
    for (uint16_t w = startWord; w < endWord; w++) {
        // Read the value
        simulatorRequestWord(w, word, 32);
        for (uint8_t cell = 0; cell < 32; cell++) {
            if (word[cell] == 50) continue;
            if (word[cell] < SEARCH_WATCH_VALUE) {
                //printf("watch\n");
                watchRoutine(cell + w * 32, word[cell]);
            } else if (word[cell] < SEARCH_ERROR_VALUE) {
                printf("error\n");
                // Do error routine
            }
        }
    }
}

uint8_t pruneCellSimple(uint32_t cell) {
    dataPoint *p = storageGetFirstDataPoint(cell);
    if (p == NULL) {
        // There does not exist a cell
        return 0;
    }
    dataPoint *prev;
    uint8_t nmbOfRepeats = 0;
    uint8_t nmbOfPrunes = 0;
    while (p->data.next != 0) {
        prev = p;
        p = storageGetNextPtr(p);
        if (prev->data.meas == p->data.meas) {
            nmbOfRepeats++;
            if (nmbOfRepeats > SIMPLE_PRUNE_LIMIT) {
                nmbOfRepeats = 0;
                nmbOfPrunes++;
                normalPrune++;
                printf("Pruned cell %d at time %d\n", cell, prev->data.time);
                datalogWritePrunedCell(cell, 0);
                assert(storageDeleteDatapoint(cell, prev->data.time) == 1);
            }
        }
    }
    return nmbOfPrunes;
}

uint8_t pruneCellErratic(uint32_t cell){
    // Calculate deltaT
    dataPoint *first = storageGetFirstDataPoint(cell);
    dataPoint *last = storageGetLatestDataPoint(cell);
    if(first == NULL || last == NULL){
        return 0;
    }

    uint16_t deltaT = last->data.time - first->data.time;

    if(deltaT > PRUNE_ERRATIC_DELTA_T_THRESHOLD && totalDerivative(cell) > PRUNE_ERRATIC_TD_THRESHOLD){
        storageDeleteCell(cell);
        printf("Pruned cell %d with deltaT = %d, td = %f\n",cell, deltaT, totalDerivative(cell));
        datalogWritePrunedCell(cell, 1);
        erraticPrune++;
        return 1;
    }

    return 0;
}

double cellValue(uint32_t cell) {
    /*
     * Make sure the cell has datapoints
    */
    if (storageGetNmbOfDataPoints(cell) == 0) {
        return 0;
    }
    // Two factors, abs value and trend.
    double d3, td;
    d3 = -der3(cell);
    td = -totalDerivative(cell);
    int16_t absChange = absoluteChange(cell);
    return d3 * C_TREND_SHORT_WEIGHT + td * C_TREND_LONG_WEIGHT + (double) absChange * C_ABSOLUTE_WEIGHT;
}

double datapointValue(dataPoint *p, uint32_t cell, uint8_t isNewCell) {
    if (p == storageGetFirstDataPoint(cell)) {
        // First point, never delete this. Infinte value
        return INFINITY;
    }

    if (isNewCell) {
        // Cell is new, infinite value
        return INFINITY;
    }

    // Distance to neighbouring cells
    dataPoint *prev;
    // storageGetPrvPtr assumes the cell you are evaluating on is stored. If this is a new cell not yet stored,
    // the prev ptr would be the last ptr stored.
    if (isNewCell) {
        prev = storageGetLatestDataPoint(cell);
    } else {
        prev = storageGetPrvPtr(p, cell);
    }
    dataPoint *next = storageGetNextPtr(p);

    uint16_t prevDistance, nextDistance;
    uint16_t prevDelta, nextDelta;

    if (prev != NULL) {
        // There a previous ptr
        prevDistance = (uint16_t) abs(p->data.time - prev->data.time);
        prevDelta = (uint16_t) abs(p->data.meas - prev->data.meas);
    }

    if (next != NULL) {
        nextDistance = (uint16_t) abs(next->data.time - p->data.time);
        nextDelta = (uint16_t) abs(p->data.meas - next->data.meas);
    }

    // Handle edge cases when prev == null or next == null
    if (prev == NULL) {
        if (next != NULL) {
            prevDistance = nextDistance;
            prevDelta = nextDelta;
        } else {
            prevDistance = prevDelta = 0;
        }
    }

    if (next == NULL) {
        if (prev != NULL) {
            nextDistance = prevDistance;
            nextDelta = prevDelta;
        } else {
            nextDistance = nextDelta = 0;
        }
    }

    uint16_t distance = prevDistance + nextDistance;

    // Delta I with neighbours
    uint16_t deltaI = prevDelta + nextDelta;

    // Distance to edge
    dataPoint *first = storageGetFirstDataPoint(cell);
    dataPoint *last = storageGetLatestDataPoint(cell);
    double t = (double) p->data.time / (double) (first->data.time + last->data.time) - 0.5;
    double distanceEdge = pow(t, 2);

    // Make sure distance does not dominate
    if (distance > 500) distance = 500;

    return DP_DELTA_WEIGHT * deltaI + DP_DISTANCE_WEIGHT * distance + DP_DISTANCE_EDGE_WEIGHT * distanceEdge;
}

// https://www.geeksforgeeks.org/union-and-intersection-of-two-sorted-arrays-2/
int arrayUnion(uint32_t *arr1, uint32_t *arr2, uint32_t *out, int m, int n, int s) {
    int i = 0, j = 0, k = 0;
    while (i < m && j < n) {
        if (arr1[i] < arr2[j]) {
            if(k >= s) return 0;
            out[k] = arr1[i++];
            k++;
        } else if (arr2[j] < arr1[i]) {
            if(k >= s) return 0;
            out[k] = arr2[j++];
            k++;
        } else {
            if(k >= s) return 0;
            out[k] = arr2[j++];
            k++;
            i++;
        }
    }
    while (i < m) {
        if(k >= s) return 0;
        out[k] = arr1[i++];
        k++;
    }
    while (j < n) {
        if(k >= s) return 0;
        out[k] = arr2[j++];
        k++;
    }
    return k;
}

void reprogramCell(uint32_t cell){
    simulatorProgram(cell);
    storageDeleteCell(cell);
}

uint8_t evaluateCellForReprogramming(uint32_t cell){
    dataPoint *last = storageGetLatestDataPoint(cell);
    if(last == NULL){
        return 0;
    }

    // Need at least 3 points to have a big enough picture
    if(storageGetNmbOfDataPoints(cell) < 3){
        return 0;
    }

    // Cell must be lower than warning absolute
    if(last->data.meas > REPROGRAM_WARNING_ABSOLUTE){
        return 0;
    }

    // Is the cell lower than the critical, reprogram
    if(last->data.meas < REPROGRAM_CIRITCAL_ABSOLUTE){
        // Cell has a very low absolute value. Reprogram
        reprogramCell(cell);
        return 1;
    }

    // Cell is lower than the warning but higher than the critical. Check derivatives
    double td = totalDerivative(cell);

    // Find DELTA
    dataPoint *points[3];
    points[2] = last;
    points[1] = storageGetPrvPtr(last, cell);
    points[0] = storageGetPrvPtr(points[1], cell);

    uint16_t max = 0;
    uint16_t min = 0xFFFF;
    for(int i = 0; i < REPGORAM_DELTA_POINT_BASE; i++){
        if(points[i]->data.meas > max) max = points[i]->data.meas;
        if(points[i]->data.meas < min) min = points[i]->data.meas;
    }
    uint16_t delta = abs(max - min);


    if(td < REPROGRAM_CIRITCAL_TD && delta < REPGORAM_ABS_VALUE_DELTA){
        if(cell == 87){
            printf("Reprogramming error, td = %f, delta = %d\n", td, delta);
        }
        reprogramCell(cell);
        return 1;
    }

    return 0;
}

void searchEvaluateReprogramming(){
    // Get a list of all existing cells
    uint32_t existingCells[100];
    int numberOfExistingCells = storageGetListOfCells(existingCells, 100);

    for (int i = 0; i < numberOfExistingCells; ++i) {
        if(evaluateCellForReprogramming(existingCells[i]) == 1){
            printf("Reprogrammed %d\n", existingCells[i]);
        }
    }
}

void searchSweep() {
    // Do a sweep on SEARCH_WATCH_VALUE
    uint32_t sweepCells[10000];
    int nmbOfSweepCells = simulatorSweep(SEARCH_WATCH_VALUE, sweepCells, 10000);

    // Get a list of all existing cells
    uint32_t existingCells[100];
    int numberOfExistingCells = storageGetListOfCells(existingCells, 100);

    // Create a "search set" which ishy sweepCells UNION existingCells
    uint32_t searchSet[100];
    int searchSetLength = arrayUnion(sweepCells, existingCells, searchSet, nmbOfSweepCells, numberOfExistingCells, 100);

    // Do a measurement of all the cells in the search set
    uint32_t v;
    for (int j = 0; j < searchSetLength; j++) {
        simulatorRqstBit(searchSet[j], &v);
        watchRoutine(searchSet[j], v);
    }

    // Do a reprogramming evaluation of all logged cells
#ifndef NO_REPROGRAM
    searchEvaluateReprogramming();
#endif

    datalogWriteDelete(normalPrune, erraticPrune, sameDelete, otherDelete);
}

void printDebugDeleteInfo() {
    printf("Prune Simple = %d, Prune Erratic = %d, Same Delete = %d, Other Delete = %d\n", normalPrune, erraticPrune, sameDelete, otherDelete);
}

