#include "simulatorNetworkInterface.h"
#include <stdio.h>
#include <stdint.h>
#include <memory.h>
#include <sys/socket.h>
#include <netdb.h>
#include <assert.h>
#include <stdlib.h>
#include <unistd.h>
#include "../jsmn/jsmn.h"

typedef struct jsonData {
    char status[200];
    uint16_t sim_nmb;
    uint32_t *data;
    uint16_t dataLength;
} jsonData;

int sockfd = -1;
static int simulationNumber = 241;

void setSimNmb(int simNmb){
    simulationNumber = simNmb;
}


static int jsoneq(const char *json, jsmntok_t *tok, const char *s) {
    if (tok->type == JSMN_STRING && (int) strlen(s) == tok->end - tok->start &&
        strncmp(json + tok->start, s, tok->end - tok->start) == 0) {
        return 0;
    }
    return -1;
}

void parseJson(struct jsonData *jsonStruct, char *jsonString, int jsonLen) {
    jsmn_parser p;
    jsmntok_t t[128];

    jsmn_init(&p);
    int r = jsmn_parse(&p, jsonString, strlen(jsonString), t, sizeof(t) / sizeof(t[0]));
    if (r < 0) {
        printf("Failed to parse JSON: %d\n", r);
        return;
    }

    /* Assume the top-level element is an object */
    if (r < 1 || t[0].type != JSMN_OBJECT) {
        printf("Object expected\n");
        return;
    }

    /* Loop over all keys of the root object */
    for (int i = 1; i < r; i++) {
        if (jsoneq(jsonString, &t[i], "status") == 0) {
            memcpy(jsonStruct->status, jsonString + t[i + 1].start, t[i + 1].end - t[i + 1].start);
        } else if (jsoneq(jsonString, &t[i], "sim_nmb") == 0) {
            jsonStruct->sim_nmb = (uint16_t) strtol(jsonString + t[i + 1].start, NULL, 10);
        } else if (jsoneq(jsonString, &t[i], "data") == 0) {
            if (t[i + 1].type != JSMN_ARRAY) {
                continue;
            }
            // Make sure that the data is allocated
            assert(jsonStruct->dataLength != 0);
            jsonStruct->dataLength = 0;
            char buffer[strlen(jsonString)];
            for (int j = 0; j < t[i + 1].size; j++) {
                jsmntok_t *g = &t[i + j + 2];
                memcpy(buffer, jsonString + g->start, g->end - g->start);
                buffer[g->end - g->start] = '\0';
                jsonStruct->data[jsonStruct->dataLength] = (uint32_t) atoi(buffer);
                jsonStruct->dataLength++;
            }
            i += t[i + 1].size + 1;
        }
    }
}

void simulatorConnect() {
    // Use getaddrinfo to populate res with options hints
    struct addrinfo hints, *res;
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    getaddrinfo("127.0.0.1", "8181", &hints, &res);

    // Make a socket with the results
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    assert(sockfd != -1);

    // Connect
    assert(connect(sockfd, res->ai_addr, res->ai_addrlen) != -1);

    freeaddrinfo(res);
}

void simulatorStart() {
    simulatorConnect();

    // Prepare start_sim command
    char *msg = "{\"cmd\": \"start_sim\"}";
    assert(send(sockfd, msg, strlen(msg), 0) != -1);

    // Receive JSON string
    char buf[300];
    assert(recv(sockfd, buf, 300 * sizeof(char), 0) != -1);
#ifdef VERBOSE
    printf("%s\n", receivedJson);
#endif

    // Prepare JSON Struct and parse it
    jsonData j = {};
    uint32_t data[100];
    j.data = data;
    j.dataLength = 100;
    parseJson(&j, buf, strlen(buf));

    // Assert that everything is ok
    assert(strcmp(j.status, "ok") == 0);

    // Store simulationNumber
    simulationNumber = j.sim_nmb;

    close(sockfd);

    printf("Sim Nmb: %d\n", simulationNumber);

}

void simulatorProgressTime(int n) {
    simulatorConnect();

    // Prepare and send command string
    assert(simulationNumber != -1);
    char msg[strlen("{\"cmd\": \"prgs_time\", \"sim_nmb\": 0000, \"n\": 0000}")];
    sprintf(msg, "{\"cmd\": \"prgs_time\", \"sim_nmb\": %d, \"n\": %d}", simulationNumber, n);
    assert(send(sockfd, msg, strlen(msg), 0) != -1);

    // Receive the JSON
    char jsonString[300];
    memset(jsonString, 0, 300);
    assert(recv(sockfd, jsonString, 300 * sizeof(char), 0) != -1);
#ifdef VERBOSE
    printf("%s\n", receivedJson);
#endif
    // Prepare JSON Struct and parse it
    jsonData j = {};
    parseJson(&j, jsonString, strlen(jsonString));

    // Assert that everything is ok
    assert(strcmp(j.status, "ok") == 0);

    close(sockfd);
}

/**
 * Fetch the icell measurement for a word from the simulator.
 * @param word Word to be fetched
 * @param data Preallocated array for the icell values
 * @param dataLength Length of the preallocated array data
 */
void simulatorRequestWord(uint16_t word, uint32_t *data, uint16_t dataLength) {
    simulatorConnect();

    // Prepare and send command string
    assert(simulationNumber != -1);
    char msg[strlen("{\"cmd\": \"rqst_data_word\", \"sim_nmb\": 0000, \"word\": 0000}")];
    sprintf(msg, "{\"cmd\": \"rqst_data_word\", \"sim_nmb\": %d, \"word\": %d}", simulationNumber, word);
    assert(send(sockfd, msg, strlen(msg), 0) != -1);

    // Receive the JSON
    char receivedJson[1000];
    memset(receivedJson, 0, 1000);
    assert(recv(sockfd, receivedJson, 1000 * sizeof(char), 0) != -1);
#ifdef VERBOSE
    printf("JSON : %s\n", receivedJson);
#endif

    // Prepare JSON Struct and parse it
    jsonData j = {};
    j.data = data;
    j.dataLength = dataLength;

#ifdef VERBOSE
    printf("%s\n", receivedJson);
#endif
    parseJson(&j, receivedJson, strlen(receivedJson));

    // Assert that everything is ok
    assert(strcmp(j.status, "ok") == 0);

    close(sockfd);
}

void simulatorRqstBit(uint32_t bit, uint32_t *data) {
    simulatorConnect();

    // Prepare and send command string
    assert(simulationNumber != -1);
    char msg[strlen("{\"cmd\": \"rqst_data_sb\", \"sim_nmb\": 0000, \"bit\": 0000}")];
    sprintf(msg, "{\"cmd\": \"rqst_data_sb\", \"sim_nmb\": %d, \"bit\": %d}", simulationNumber, bit);
    assert(send(sockfd, msg, strlen(msg), 0) != -1);

    // Receive the JSON
    char receivedJson[1000];
    memset(receivedJson, 0, 1000);
    assert(recv(sockfd, receivedJson, 1000 * sizeof(char), 0) != -1);
#ifdef VERBOSE
    printf("JSON : %s\n", receivedJson);
#endif

    // Prepare JSON Struct and parse it
    jsonData j = {};
    j.data = data;
    j.dataLength = 1;

#ifdef VERBOSE
    printf("%s\n", receivedJson);
#endif
    parseJson(&j, receivedJson, strlen(receivedJson));

    // Assert that everything is ok
    assert(strcmp(j.status, "ok") == 0);

    close(sockfd);
}

uint16_t simulatorRequestTime(){
    simulatorConnect();

    // Prepare and send command string
    assert(simulationNumber != -1);
    char msg[strlen("{\"cmd\": \"rqst_time\", \"sim_nmb\": 0000}")];
    sprintf(msg, "{\"cmd\": \"rqst_time\", \"sim_nmb\": %d}", simulationNumber);
    assert(send(sockfd, msg, strlen(msg), 0) != -1);

    // Receive the JSON
    char receivedJson[300];
    memset(receivedJson, 0, 300);
    assert(recv(sockfd, receivedJson, 300 * sizeof(char), 0) != -1);
#ifdef VERBOSE
    printf("JSON : %s\n", receivedJson);
#endif

    // Prepare JSON Struct and parse it
    uint32_t data[1];
    jsonData j = {};
    j.data = data;
    j.dataLength = 1;
#ifdef VERBOSE
    printf("%s\n", receivedJson);
#endif
    parseJson(&j, receivedJson, strlen(receivedJson));

    // Assert that everything is ok
    assert(strcmp(j.status, "ok") == 0);

    close(sockfd);

    return j.data[0];
}

int simulatorSweep(uint16_t threshold, uint32_t *data, uint16_t dataLength) {
    simulatorConnect();

    // Prepare and send command string
    assert(simulationNumber != -1);
    char msg[strlen("{\"cmd\": \"sweep\", \"sim_nmb\": 0000, \"threshold\": 0000}")];
    sprintf(msg, "{\"cmd\": \"sweep\", \"sim_nmb\": %d, \"threshold\": %d}", simulationNumber, threshold);
    assert(send(sockfd, msg, strlen(msg), 0) != -1);

    // Receive the JSON
    char receivedJson[1000];
    memset(receivedJson, 0, 1000);
    assert(recv(sockfd, receivedJson, 1000 * sizeof(char), 0) != -1);
#ifdef VERBOSE
    printf("JSON : %s\n", receivedJson);
#endif

    // Prepare JSON Struct and parse it
    jsonData j = {};
    j.data = data;
    j.dataLength = dataLength;

#ifdef VERBOSE
    printf("%s\n", receivedJson);
#endif
    parseJson(&j, receivedJson, strlen(receivedJson));

    // Assert that everything is ok
    assert(strcmp(j.status, "ok") == 0);

    close(sockfd);

    return j.dataLength;
}

void simulatorProgram(uint32_t bit) {
    simulatorConnect();

    // Prepare and send command string
    assert(simulationNumber != -1);
    char msg[strlen("{\"cmd\": \"program\", \"sim_nmb\": 0000, \"bit\": 0000}")];
    sprintf(msg, "{\"cmd\": \"program\", \"sim_nmb\": %d, \"bit\": %d}", simulationNumber, bit);
    assert(send(sockfd, msg, strlen(msg), 0) != -1);

    // Receive the JSON
    char jsonString[300];
    memset(jsonString, 0, 300);
    assert(recv(sockfd, jsonString, 300 * sizeof(char), 0) != -1);
#ifdef VERBOSE
    printf("%s\n", receivedJson);
#endif
    // Prepare JSON Struct and parse it
    jsonData j = {};
    parseJson(&j, jsonString, strlen(jsonString));

    // Assert that everything is ok
    assert(strcmp(j.status, "ok") == 0);

    close(sockfd);
}