#include "statistics.h"
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <stdio.h>
#include <memory.h>
#include "../simulatorNetworkInterface/simulatorNetworkInterface.h"

static double calibratedMean = 22.0;

/**
 * Searches for an element in an array. Linear search O(N)
 * @param element Array to find
 * @param array Array to search in
 * @param arrayLength Length of array to search in
 * @return 1 if element exists in array, 0 if not
 */
int elementInArray(uint32_t element, uint32_t const *array, uint32_t arrayLength) {
    for (uint32_t i = 0; i < arrayLength; i++) {
        if (element == array[i]) return 1;
    }
    return 0;
}

/**
 * RobertFloyd random selection implementation. Selects M distinct elements from the range [0..N]
 * Source: https://stackoverflow.com/questions/2394246/algorithm-to-select-a-single-random-combination-of-values
 *
 * @param N The upper range of selection where the range is [0..N]
 * @param M The number of distinct elements to select
 * @param S The result of the random selection of size M
 */
void randomSelectMFromZeroToN(uint32_t N, uint32_t M, uint32_t *S) {
    srand(time(NULL));
    int sI = 0;
    memset(S, 0, M);

    uint32_t T;
    for (uint32_t j = N - M + 1; j < N; j++) {
        T = (uint32_t) rand() % j;
        if (!elementInArray(T, S, M)) {
            S[sI] = T;
            sI++;
        } else {
            S[sI] = j;
            sI++;
        }
    }
}

/**
 * Calculates the mean on data with the central limit theorem with a confidence
 * @param data Data to calculate mean on
 * @param size Size of data
 * @param confidence The size of the subset to calculate the mean on
 * @return Estimated mean
 */
double calculateMeanCentralLimit(uint32_t *data, uint32_t size, uint32_t confidence) {
    uint32_t randIndexes[confidence];
    randomSelectMFromZeroToN(size, confidence, randIndexes);

    uint32_t sum = 0;
    for (uint32_t i = 0; i < confidence; i++) {
        sum += data[randIndexes[i]];
    }

    return (double) sum / (double) confidence;
}

/**
 * Calculates the mean of the icell distribution on the flash using the central limit theorem.
 * @param nmbOfWords The number of words the mean should be calcualted on. The higher the number, the more
 * accurate the result is. If nmbOfWords is equal the total number of words in the flash, the accuracy is 100%
 * @return The calculated mean
 */
double calculateMeanCentralLimitFlash(uint32_t nmbOfWords) {
    // Select M distinct words
    uint32_t randIndexes[nmbOfWords];
    randomSelectMFromZeroToN(FLASH_SIZE_KB * 1024 / 4, nmbOfWords, randIndexes);
    uint16_t data[nmbOfWords * 32];
    for (int w = 0; w < nmbOfWords; w++) {
        simulatorRequestWord(randIndexes[w], data + w * 32, 32);
    }

    double sum = 0;
    int nmbOfErased = 0;
    int k;
    for (int w = 0; w < nmbOfWords; w++) {
        for (int b = 0; b < 32; b++) {
            k = data[w * 32 + b];
            if (k != 50) {
                sum += k;
                nmbOfErased++;
            }
        }
    }

    return sum / nmbOfErased;
}

/**
 * Calculates the temperature error induced by using an already calibrated mean
 * @param confidence The higher the confidence, the higher the accuracy
 * @return The temperature induced error
 */
double temperatureError(uint32_t confidence) {
    return calculateMeanCentralLimitFlash(confidence) - calibratedMean;
}