#include "storage.h"
#include <memory.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <assert.h>
#include "../search/search.h"

int NMB_OF_BLOCKS;
int MAX_NMB_OF_CELLS;
int PREFERED_NMB_OF_CELLS;
int MAX_NMB_OF_POINTS_FOR_CELL;

dataPoint *nvmModel = NULL;
uint8_t *start;
uint32_t *cellIndex;
uint32_t *repgorammedCells;

dataPoint *getModel() {
    return nvmModel;
}

void initializeDebugData() {
    for (int i = 0; i < NMB_OF_BLOCKS; i++) {
        nvmModel[i].data.metadata.DEBUG_DATA = 0;
    }
}

void storageInitNVMModel(uint16_t nmbOfBlocks, uint16_t maxNmbOfCells, uint16_t maxNmbPtsPerCell,
                         uint16_t preferedNmbOfCells) {
    NMB_OF_BLOCKS = nmbOfBlocks;
    MAX_NMB_OF_CELLS = maxNmbOfCells;
    MAX_NMB_OF_POINTS_FOR_CELL = maxNmbPtsPerCell;
    PREFERED_NMB_OF_CELLS = preferedNmbOfCells;

    start = malloc(sizeof(start) * MAX_NMB_OF_CELLS);
    cellIndex = malloc(sizeof(cellIndex) * MAX_NMB_OF_CELLS);
    repgorammedCells = malloc(sizeof(repgorammedCells) * PREFERED_NMB_OF_CELLS);

    nvmModel = malloc(sizeof(nvmModel) * NMB_OF_BLOCKS);
    for (int i = 0; i < NMB_OF_BLOCKS; i++) {
        nvmModel[i].DATA = 0;
    }
    for (int i = 0; i < MAX_NMB_OF_CELLS; i++) {
        cellIndex[i] = CELL_INDEX_EMPTY;
        start[i] = 0;
    }
    initializeDebugData();
}

void storageFreeNVMModel() {
    free(nvmModel);
    free(start);
    free(cellIndex);
}

uint8_t getIndexOfCell(const uint32_t cell) {
    for (uint8_t i = 0; i < MAX_NMB_OF_CELLS; i++) {
        if (cellIndex[i] == cell) return i;
    }
    return ERR_CELL_INDEX_NOT_EXIST;
}

uint8_t addCellIndex(const uint32_t cell) {
    for (uint8_t i = 0; i < MAX_NMB_OF_CELLS; i++) {
        if (cellIndex[i] == CELL_INDEX_EMPTY) {
            cellIndex[i] = cell;
            return i;
        }
    }
    assert(1==0);
    return ERR_CELL_INDEX_NOT_EXIST;
}

dataPoint *storageGetNextPtr(struct dataPoint *p) {
    if (p == NULL) {
        printf("Warning, uninitialized datapoint to storageGetNextPtr");
        return NULL;
    }
    if (p->data.next == 0) {
        return NULL;
    }
    return (p->data.next + nvmModel);
}

dataPoint *storageGetPrvPtr(struct dataPoint *p, uint32_t cell) {
    dataPoint *next = storageGetFirstDataPoint(cell);

    // Is there no previous points? Return NULL
    if (next == p) {
        return NULL;
    }

    // Start at the beginning. Stop when p is found
    while (next->data.next != (p - nvmModel)) {
        next = storageGetNextPtr(next);
    }
    return next;
}

dataPoint *getFirstPtr(const uint8_t index) {
    return (start[index] + nvmModel);
}

dataPoint *findLatestDatapoint(struct dataPoint *p) {
    struct dataPoint *next = p;
    while (next->data.next != 0) {
        next = storageGetNextPtr(next);
    }
    return next;
}

uint8_t findEmptyBlock() {
    for (uint8_t i = 0; i < NMB_OF_BLOCKS; i++) {
        if (nvmModel[i].data.time == 0) return i;
    }
    return ERR_NO_MORE_EMPTY_BLOCKS;
}

int storageAddDatapoint(struct dataPoint *p, uint32_t cell) {
    uint8_t cIndex = getIndexOfCell(cell);
    uint8_t block = findEmptyBlock();
    // Update debug data
    p->data.metadata.DEBUG_DATA = nvmModel[block].data.metadata.DEBUG_DATA; // Carry over debug data
    p->data.metadata.dData.nmbOfWrites++;
    // Make sure next ptr is 0
    p->data.next = 0;

    if (cIndex != ERR_CELL_INDEX_NOT_EXIST) {
        // There already exists a datapoint, find the latest datapoint
        struct dataPoint *last = findLatestDatapoint(getFirstPtr(cIndex));
        // Find an empty block
        if (block == ERR_NO_MORE_EMPTY_BLOCKS) {
            return 0;
        }
        // Make sure block is empty. Sanity check
        assert(nvmModel[block].data.time == 0);

        // Add the new datapoint
        nvmModel[block] = *p;
        last->data.next = block;
        return 1;
    } else {
        // There does not already exist a datapoint
        // Add the new datapoint
        nvmModel[block] = *p;
        start[addCellIndex(cell)] = block;
        return 1;
    }
}

uint8_t storageDeleteDatapoint(uint32_t cell, uint16_t time) {
    // Find datapoint with time == time
    dataPoint *p = storageGetFirstDataPoint(cell);
    if (p == NULL) {
        // There does not exists a datapoint for this cell
        return 0;
    }
    dataPoint *prev = NULL;
    while (p->data.time != time) {
        if (p->data.time == 0) {
            // Datapoint with time t does not exist
            return 0;
        }
        prev = p;
        p = storageGetNextPtr(p);
        if (p == NULL) {
            // Next ptr does not exist
            return 0;
        }
    }
    // Update metadata
    p->data.metadata.dData.nmbOfDeletes++;

    // Rearrange next ptrs
    if (prev != NULL) {
        prev->data.next = p->data.next;
    } else {
        // No previous ptr exist, must update start ptr
        start[getIndexOfCell(cell)] = (uint8_t) (p - nvmModel);
    }
    // Clear datapoint
    p->data.time = p->data.meas = p->data.next = 0;

    if (storageGetNmbOfDataPoints(cell) == 0) {
        // Last point, remove from index
        cellIndex[getIndexOfCell(cell)] = ERR_CELL_INDEX_NOT_EXIST;
    }
    return 1;

}

int storageDeleteCell(uint32_t cell) {
    dataPoint *p = storageGetFirstDataPoint(cell);
    if (p == NULL) {
        return 0;
    }
    dataPoint *next;
    while (p->data.next != 0) {
        next = storageGetNextPtr(p);
        if (next == NULL) {
            return 0;
        }
        // Delete point
        p->data.time = p->data.meas = p->data.next = 0;
        // Update metadata
        p->data.metadata.dData.nmbOfDeletes++;
        p = next;
    }
    // Delete point
    p->data.time = p->data.meas = p->data.next = 0;
    // Update metadata
    p->data.metadata.dData.nmbOfDeletes++;

    // Delete from cell index
    cellIndex[getIndexOfCell(cell)] = CELL_INDEX_EMPTY;
    return 1;
}

uint8_t storageDoesDatapointsExist(uint32_t cell) {
    return (uint8_t) (getIndexOfCell(cell) != ERR_CELL_INDEX_NOT_EXIST);
}

dataPoint *storageGetLatestDataPoint(uint32_t cell) {
    uint8_t index = getIndexOfCell(cell);
    if (index == ERR_CELL_INDEX_NOT_EXIST) {
        return NULL;
    }
    return findLatestDatapoint(getFirstPtr(index));
}

dataPoint *storageGetFirstDataPoint(uint32_t cell) {
    uint8_t index = getIndexOfCell(cell);
    if (index == ERR_CELL_INDEX_NOT_EXIST) {
        return NULL;
    }
    return getFirstPtr(index);
}

dataPoint *storageGetDatapointNmbX(uint32_t cell, uint8_t X) {
    uint8_t index = getIndexOfCell(cell);
    if (index == ERR_CELL_INDEX_NOT_EXIST) {
        return NULL;
    }
    struct dataPoint *next = getFirstPtr(index);
    for (int i = 0; i < X; i++) {
        next = storageGetNextPtr(next);
        if (next->data.next == 0) {
            if (next->data.time == 0) {
                printf("Error in storageGetDatapointNmbX");
                exit(23);
            } else {
                return next;
            }
        }
    }
    return next;
}

uint8_t storageGetNmbOfDataPoints(uint32_t cell) {
    uint8_t index = getIndexOfCell(cell);
    if (index == ERR_CELL_INDEX_NOT_EXIST) {
        return 0;
    }
    struct dataPoint *next = getFirstPtr(index);

    // First datapoint does not exist, old cell index not removed.
    if (next->data.time == 0) return 0;

    uint8_t c = 1;
    while (next->data.next != 0) {
        next = storageGetNextPtr(next);
        c++;
    }
    return c;
}

uint16_t storageGetListOfCells(uint32_t *cellList, uint16_t size) {
    uint16_t j = 0;
    for (int i = 0; i < MAX_NMB_OF_CELLS; i++) {
        if (cellIndex[i] != CELL_INDEX_EMPTY) {
            if (j > size) {
                return 0;
            }
            cellList[j] = cellIndex[i];
            j++;
        }
    }
    return j;
}

uint8_t storageGetNmbEmptyBlocks() {
    uint8_t nmbOfOccupiedBlocks = 0;
    dataPoint *p;
    for (int i = 0; i < MAX_NMB_OF_CELLS; i++) {
        if (cellIndex[i] != CELL_INDEX_EMPTY) {
            // There exists a cell here
            p = getFirstPtr(i);
            assert(p != NULL);
            nmbOfOccupiedBlocks++;
            while (p->data.next != 0) {
                nmbOfOccupiedBlocks++;
                p = storageGetNextPtr(p);
            }
        }
    }
    return (uint8_t) NMB_OF_BLOCKS - nmbOfOccupiedBlocks - 1;
}

uint8_t storageGetNmbOfCells() {
    uint8_t nmbOfCells = 0;
    for (int i = 0; i < MAX_NMB_OF_CELLS; i++) {
        if (cellIndex[i] != CELL_INDEX_EMPTY) {
            nmbOfCells++;
        }
    }
    return nmbOfCells;
}

double storageFindLowestValuedDatapoint(uint32_t cell, dataPoint **lowestPoint) {
    double lowestFound = INFINITY;
    double value = 0;
    dataPoint *p;
    uint8_t i = getIndexOfCell(cell);
    if (i != ERR_CELL_INDEX_NOT_EXIST) {
        // There exists a cell here
        p = getFirstPtr(i);
        assert(p != NULL);

        uint8_t morePointsToProcess = 1;
        while (morePointsToProcess) {
            if (p->data.next == 0) {
                morePointsToProcess = 0;
            }
            value = datapointValue(p, cell, 0);
            if (value <= lowestFound) {
                lowestFound = value;
                *lowestPoint = p;
            }
            p = storageGetNextPtr(p);
        }
    } else {
        printf("Warning no cell index\n");
        exit(23);
    }
    return lowestFound;
}

void storageFindLowestValuedCell(uint32_t *cell, double *value) {
    *value = INFINITY;
    uint32_t c;
    double v;
    for (int i = 0; i < MAX_NMB_OF_CELLS; i++) {
        c = cellIndex[i];
        if (c != CELL_INDEX_EMPTY) {
            // There exists a cell here
            v = cellValue(c);
            if (v < *value) {
                *value = v;
                *cell = c;
            }
        }
    }
}

uint8_t storageFindCellWithMostPoints(uint32_t *cell, uint8_t *nmbOfPoints) {
    dataPoint *p;
    uint8_t min = 0;
    uint32_t cellWithMin;

    // Does there exist any?
    if (storageGetNmbEmptyBlocks() == NMB_OF_BLOCKS) return 0;

    for (int i = 0; i < MAX_NMB_OF_CELLS; i++) {
        *cell = cellIndex[i];
        if (cellIndex[i] != CELL_INDEX_EMPTY) {
            // There exists a cell here
            p = getFirstPtr(i);
            assert(p != NULL);
            *nmbOfPoints = storageGetNmbOfDataPoints(*cell);
            if (*nmbOfPoints > min) {
                min = *nmbOfPoints;
                cellWithMin = *cell;
            }
        }
    }

    *cell = cellWithMin;
    *nmbOfPoints = min;
    return 1;
}
