#ifndef PREVENTIONSYSTEM_STORAGE_H
#define PREVENTIONSYSTEM_STORAGE_H
#include <stdint.h>

#define DEDICATED_FLASH_SIZE 1024

#define CELL_INDEX_EMPTY (FLASH_SIZE_KB*1024*8 + 1)
#define ERR_CELL_INDEX_NOT_EXIST (MAX_NMB_OF_CELLS + 1)
#define ERR_NO_MORE_EMPTY_BLOCKS (7*32 + 31)


#ifdef __cplusplus
extern "C" {
#endif

/**
 * Variable defining number of datapoints that can be allocted by the system
 */
extern int NMB_OF_BLOCKS;

/**
 * The maximum number of cells the system can have
 */
extern int MAX_NMB_OF_CELLS;

/**
 * Prefered max number of cells to have in the system
 */
extern int PREFERED_NMB_OF_CELLS;

/**
 * Maximum number of points for any given cell
 */
extern int MAX_NMB_OF_POINTS_FOR_CELL;

typedef struct debug_data {
    uint16_t nmbOfDeletes;
    uint16_t nmbOfWrites;
} debug_data;

typedef struct debugData {
    union {
        uint32_t DEBUG_DATA;
        debug_data dData;
    };
} debugData;

typedef struct dataPointData {
    unsigned int meas : 12;
    unsigned int time : 12;
    uint8_t next : 8;
    debugData metadata;
} dataPointData;

typedef struct dataPoint {
    union {
        uint64_t DATA;
        dataPointData data;
    };
} dataPoint;

/**
 * Initializes the NVM Model, allocating the memory
 * @param nmbOfBlocks Number of blocks to have in the system. One block can hold one datapoint
 * @param maxNmbOfCells Maximum number of cells allowed in the system
 * @param maxNmbPtsPerCell Maximum number of points any given cell can obtain
 * @param preferedNmbOfCells Preferd maximum number of points any given cell can obtain. This differs from
 * the maxNmbPtsPerCell on that if there are unused blocks, the cell can still takes those. However, all number of cells
 * above this number is easily lost.
 */
void storageInitNVMModel(uint16_t nmbOfBlocks, uint16_t maxNmbOfCells, uint16_t maxNmbPtsPerCell, uint16_t preferedNmbOfCells);

/**
 * Frees the NVM Model from the heap allocated by storageInitNVMModel()
 */
void storageFreeNVMModel();

/**
 * Adds a datapoint to the NVM Model
 * @param p Datapoint to add
 * @param cell Cell the datapoint belongs to
 * @return 1 on success, 0 on failure
 */
int storageAddDatapoint(struct dataPoint *p, uint32_t cell);

/**
 * Checks if at least one datapoint exists for the given cell
 * @param cell Cell to check
 * @return 1 if there exists a datapoint, 0 if not
 */
uint8_t storageDoesDatapointsExist(uint32_t cell);

/**
 * Finds the latest datapoint in time for the given cell
 * @param cell Cell to find datapoint for
 * @return Latest datapoint if it exsits, otherwise NULL
 */
dataPoint *storageGetLatestDataPoint(uint32_t cell);

/**
 * Finds the first datapoint in time for the given cell
 * @param cell Cell to find datapoint for
 * @return First datapoint if it exists, otherwise NULL
 */
dataPoint *storageGetFirstDataPoint(uint32_t cell);

/**
 * Finds the total number of datapoints for a given cell
 * @param cell Cell to find number of datapoints for
 * @return Number of datapoints for cell
 */
uint8_t storageGetNmbOfDataPoints(uint32_t cell);

/**
 * Finds datapoint number X for a given cell.
 * @param cell Cell to find datapoint for
 * @param X Datapoint number
 * @return Datapoint number X if it exists, otherwise NULL
 */
dataPoint *storageGetDatapointNmbX(uint32_t cell, uint8_t X);

/**
 * Gets the pointer for the next datapoint for a datapoint
 * @param p Datapoint to find the next pointer for
 * @return Pointer to p's next datapoint if it exists, otherwise NULL
 */
dataPoint *storageGetNextPtr(struct dataPoint *p);

/**
 * Deletes a datapoint at time for a cell
 * @param cell Cell datapoint belongs to
 * @param time Time of target datapoint
 * @return 1 if success, 0 if error
 */
uint8_t storageDeleteDatapoint(uint32_t cell, uint16_t time);

/**
 * Finds the previous datapoint in NVM Model for the datapoint
 * @param p Datapoint to find previous for
 * @param cell Cell the datapoint belongs to
 * @return Pointer to previous datapoint if it exists, NULL otherwise
 */
dataPoint *storageGetPrvPtr(struct dataPoint *p, uint32_t cell);

/**
 * Calculates the number of empty blocks
 * @return The number of empty blocks
 */
uint8_t storageGetNmbEmptyBlocks();

/**
 * Gets a list of all cells currently stored in the system
 * @param cellList The cell list is written here
 * @param size Size of the cellList array
 * @return The number of cells in the system
 */
uint16_t storageGetListOfCells(uint32_t *cellList, uint16_t size);

/**
 * Finds the lowest valued datapoint of the given cell
 * @param cell The cell to search for the lowest datapoint
 * @param lowestPoint The result is written here
 * @return The datapoint value of the lowest point
 */
double storageFindLowestValuedDatapoint(uint32_t cell, dataPoint **lowestPoint);

/**
 * Finds the cell with the most points stored
 * @param cell The cell with the most points is written here
 * @param nmbOfPoints The number of points the given cell has is written here
 * @return 1 on success, 0 on failure
 */
uint8_t storageFindCellWithMostPoints(uint32_t *cell, uint8_t *nmbOfPoints);

/**
 * Searches the NVM Model for the lowest valued cell
 * @param cell Resulting lowest valued cell is written here
 * @param value Resulting lowed valued cells value is written here
 */
void storageFindLowestValuedCell(uint32_t *cell, double *value);

/**
 * Deletes a cell and all its datapoints
 * @param cell Cell to delete
 * @return 1 on success, 0 on failure
 */
int storageDeleteCell(uint32_t cell);

/**
 * Finds the number of cells in the system
 * @return Number of cells in the system
 */
uint8_t storageGetNmbOfCells() ;

dataPoint *getModel();

#ifdef __cplusplus
}
#endif

#endif
