#include "gtest/gtest.h"
#include "storage_pi.h"

class StorageTestSuite : public ::testing::Test {
protected:
    void SetUp() override {
        storageInitNVMModel(DEDICATED_FLASH_SIZE / 4 - 1, 5, 10, 5);
    }

    void TearDown() override {
        storageFreeNVMModel();
    }

    static void generateTestData() {
        uint16_t dataMeas[] = {1800, 1800, 1750, 1750, 1700, 1650, 1650, 1650, 1550, 1550, 1550, 1550, 1600, 1600};
        uint16_t dataTime[] = {10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120, 130, 140};
        dataPoint p;
        p.DATA = 0;
        for (int i = 0; i < 14; i++) {
            p.data.time = dataTime[i];
            p.data.meas = dataMeas[i];
            storageAddDatapoint(&p, 23);
        }
    }

    static void generateTestData888(){
        uint16_t dataMeas[] = {2350, 2350, 2400, 2400, 2350};
        uint16_t dataTime[] = {727, 821, 920, 1042, 1500};
        dataPoint p;
        p.DATA = 0;
        for (int i = 0; i < 5; i++) {
            p.data.time = dataTime[i];
            p.data.meas = dataMeas[i];
            assert(storageAddDatapoint(&p, 888) == 1);
            int b = 2;
        }
    }
};

TEST_F(StorageTestSuite, addDatapoint) {
    StorageTestSuite::generateTestData();
    dataPoint *p = storageGetFirstDataPoint(23);
    ASSERT_EQ(p->data.time, 10);
    ASSERT_EQ(p->data.meas, 1800);
    p = storageGetNextPtr(p);
    ASSERT_EQ(p->data.time, 20);
    ASSERT_EQ(p->data.meas, 1800);
    p = storageGetNextPtr(p);
    p = storageGetNextPtr(p);
    p = storageGetNextPtr(p);
    ASSERT_EQ(p->data.time, 50);
    ASSERT_EQ(p->data.meas, 1700);
}

TEST_F(StorageTestSuite, deleteDatapoint) {
    StorageTestSuite::generateTestData();
    dataPoint *p = storageGetDatapointNmbX(23, 4);
    ASSERT_EQ(p->data.next, 5);
    storageDeleteDatapoint(23, 60);
    ASSERT_EQ(p->data.next, 6);
}

TEST_F(StorageTestSuite, deleteCell) {
    StorageTestSuite::generateTestData();
    StorageTestSuite::generateTestData888();
   // ASSERT_EQ(storageGetNmbOfDataPoints(23), 14);
    ASSERT_EQ(storageGetNmbOfDataPoints(888), 5);

    storageDeleteCell(23);
    uint8_t k = storageGetNmbOfDataPoints(23);
    ASSERT_EQ(storageGetNmbOfDataPoints(23), 0);
    ASSERT_EQ(storageGetNmbOfDataPoints(888), 5);
}

TEST_F(StorageTestSuite, ComplexOperationsSingleCell) {
    StorageTestSuite::generateTestData();
    storageDeleteDatapoint(23, 50);
    storageDeleteDatapoint(23, 80);

    dataPoint p;
    p.DATA = 0;
    p.data.time = 270;
    p.data.meas = 1800;
    storageAddDatapoint(&p, 23);

    storageDeleteDatapoint(23, 30);

    p.DATA = 0;
    p.data.time = 280;
    p.data.meas = 1750;
    storageAddDatapoint(&p, 23);

    // Verify integrity
    dataPoint *pv = storageGetFirstDataPoint(23);
    uint16_t verifyData[13] = {1800, 1800, 1750, 1650, 1650, 1550, 1550, 1550, 1550, 1600, 1600, 1800, 1750};
    uint16_t verifyTime[13] = {10, 20, 40, 60, 70, 90, 100, 110, 120, 130, 140, 270, 280};

    for (int i = 0; i < 13; i++) {
        ASSERT_EQ(verifyData[i], pv->data.meas);
        ASSERT_EQ(verifyTime[i], pv->data.time);
        if (i != 12) {
            pv = storageGetNextPtr(pv);
        }
    }
}

TEST_F(StorageTestSuite, MultipleCellOperations) {
    // Cell 23
    StorageTestSuite::generateTestData();

    // Cell 888
    uint16_t dataMeas[] = {2350, 2350, 2400, 2400, 2350};
    uint16_t dataTime[] = {727, 821, 920, 1042, 1500};
    dataPoint p;
    p.DATA = 0;
    for (int i = 0; i < 5; i++) {
        p.data.time = dataTime[i];
        p.data.meas = dataMeas[i];
        storageAddDatapoint(&p, 888);
    }

    // Verify both cells
    // Cell 23
    uint16_t cell23Meas[] = {1800, 1800, 1750, 1750, 1700, 1650, 1650, 1650, 1550, 1550, 1550, 1550, 1600, 1600};
    uint16_t cell23Time[] = {10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 120, 130, 140};
    dataPoint *pv = storageGetFirstDataPoint(23);
    for (int i = 0; i < 5; i++) {
        ASSERT_EQ(cell23Meas[i], pv->data.meas);
        ASSERT_EQ(cell23Time[i], pv->data.time);
        if (i != 4) {
            pv = storageGetNextPtr(pv);
        }
    }

    // Cell 888
    pv = storageGetFirstDataPoint(888);
    for (int i = 0; i < 5; i++) {
        ASSERT_EQ(dataMeas[i], pv->data.meas);
        ASSERT_EQ(dataTime[i], pv->data.time);
        if (i != 4) {
            pv = storageGetNextPtr(pv);
        }
    }
}

TEST_F(StorageTestSuite, GetPrevPtr) {
    StorageTestSuite::generateTestData();
    dataPoint *p3 = storageGetDatapointNmbX(23, 3);
    dataPoint *prev = storageGetPrvPtr(p3, 23);
    ASSERT_EQ(prev->data.time, 30);

    dataPoint *p6 = storageGetDatapointNmbX(23, 6);
    prev = storageGetPrvPtr(p6, 23);
    ASSERT_EQ(prev->data.time, 60);

    dataPoint *p0 = storageGetDatapointNmbX(23, 0);
    prev = storageGetPrvPtr(p0, 23);
    ASSERT_TRUE(prev == NULL);
}

TEST_F(StorageTestSuite, GetNmbOfEmptyBlocks) {
    StorageTestSuite::generateTestData();
    ASSERT_EQ(storageGetNmbEmptyBlocks(), 240);

    // Add more points to another cell
    uint16_t dataMeas[] = {2350, 2350, 2400, 2400, 2350};
    uint16_t dataTime[] = {727, 821, 920, 1042, 1500};
    dataPoint p;
    p.DATA = 0;
    for (int i = 0; i < 5; i++) {
        p.data.time = dataTime[i];
        p.data.meas = dataMeas[i];
        storageAddDatapoint(&p, 888);
    }

    ASSERT_EQ(storageGetNmbEmptyBlocks(), 235);
}

TEST_F(StorageTestSuite, GetNmbOfCells){
    StorageTestSuite::generateTestData();
    ASSERT_EQ(storageGetNmbOfCells(), 1);
    StorageTestSuite::generateTestData888();
    ASSERT_EQ(storageGetNmbOfCells(), 2);
    storageDeleteCell(888);
    ASSERT_EQ(storageGetNmbOfCells(), 1);
}

TEST_F(StorageTestSuite, LowestValuedDatapoint) {
    StorageTestSuite::generateTestData();
    dataPoint *p;
    storageFindLowestValuedDatapoint(23, &p);
    // TODO: Verify test
    ASSERT_EQ(p->data.time, 70);
}

TEST_F(StorageTestSuite, LowedValuedCell){
    StorageTestSuite::generateTestData();
    double value = 0;
    uint32_t cell = 0;
    storageFindLowestValuedCell(&cell, &value);
    ASSERT_EQ(cell, 23);

    // Add more points to another cell
    uint16_t dataMeas[] = {2350, 2350, 2400, 2400, 2350, 1760};
    uint16_t dataTime[] = {727, 821, 920, 1042, 1500, 1550};
    dataPoint p;
    p.DATA = 0;
    for (int i = 0; i < 6; i++) {
        p.data.time = dataTime[i];
        p.data.meas = dataMeas[i];
        storageAddDatapoint(&p, 888);
    }

    storageFindLowestValuedCell(&cell, &value);
    ASSERT_EQ(cell, 23);
}