import uuid
import numpy as np
from config import Config, ConfigErrorException
import csv
import itertools
import tqdm


class Simulation(object):
    def __init__(self):
        self.ac = Config()
        self.t = 10
        self.movingBits = np.array(self.ac.SELECTED_RANDOM_BITS, dtype=np.int64)
        self.movingBitTime = {}
        self.erraticBits = np.array(self.ac.SELECTED_ERRATIC_BITS, dtype=np.int64)
        self.activeErraticBits = np.array([], dtype=np.int64)
        self.activeErraticBitsTime = {}
        self.values = np.array([], dtype=np.int64)
        self.initialValues = np.array([], dtype=np.int64)
        self.historyChosenBits = {}
        self.interpolationB = {}
        self.interpolationC = {}
        self.bins = np.array([x * 0.25 for x in range(0, 4 * 40)])
        self.centers = (self.bins[1:] + self.bins[:-1]) / 2
        # Initialize the progression probability
        self.Pn = np.full(self.ac.FLASH_SIZE_KB * 1024 * 8, self.ac.P_PN_INITIAL)
        self.Pp = np.full(self.ac.FLASH_SIZE_KB * 1024 * 8, self.ac.P_PN_INITIAL)

        self.initializeValues()
        self.initializeErrorBits()

    def __getitem__(self, key):
        if not self.ac.REPEAT_CSV_FILES:
            return self.values[key]
        else:
            with open("{}/{}.csv".format(self.ac.CSV_FILE_DIRECTORY, self.t)) as csvfile:
                reader = csv.reader(csvfile)
                next(itertools.islice(reader, key, None))
                for row in reader:
                    return self.digitizeValue(np.float64(row[2]))

    def initializeValues(self):
        if self.ac.REPEAT_CSV_FILES:
            return

        self.values = np.zeros(self.ac.FLASH_SIZE_KB * 1024 * 8)
        # Select 50% +- 2% of the bits, mark as 0.5 muA
        # TODO: Add 2% error

        # Initialize the low current cells
        lowCurrentCells = np.append(self.ac.SELECTED_RANDOM_BITS, self.ac.SELECTED_ERRATIC_BITS)
        while np.isin(np.append(self.ac.SELECTED_RANDOM_BITS, self.ac.SELECTED_ERRATIC_BITS), lowCurrentCells).any():
            # TODO: Probably a better way to do this, as this does the calculations several times.
            lowCurrentCells = np.random.choice(self.ac.FLASH_SIZE_KB * 8 * 1024,
                                               int(self.ac.FLASH_SIZE_KB * 1024 * 8 / 2), replace=False)

        for cell in lowCurrentCells:
            self.values[cell] = 0.5

        # Initialize the high current cells
        nrOfHighCurrentCells = (self.ac.FLASH_SIZE_KB * 1024 * 8) - len(lowCurrentCells)
        highCurrentCellsValues = np.random.normal(self.ac.ICELL_MEAN, self.ac.ICELL_STD, nrOfHighCurrentCells)

        highCurrentCellIndex = 0
        for i, value in np.ndenumerate(self.values):
            if abs(value - 0.5) < 1e-10:
                continue
            self.values[i] = self.digitizeValue(highCurrentCellsValues[highCurrentCellIndex])
            highCurrentCellIndex += 1

        self.initialValues = np.copy(self.values)

    def initializeErrorBits(self):
        if self.ac.REPEAT_CSV_FILES:
            return

        # Initialize interpolation parameters
        for bit in self.movingBits:
            self.interpolationB[bit] = np.random.normal(self.ac.INTERPOLATION_B, self.ac.INTERPOLATION_B_STD, 1)[0]
            self.interpolationC[bit] = np.random.normal(self.ac.INTERPOLATION_C, self.ac.INTERPOLATION_C_STD, 1)[0]
            self.initialValues[bit] -= self.interpolationC[bit]
            self.movingBitTime[bit] = self.t

        # # Initialize history
        for bit in self.ac.HISTORY_BITS:
            self.historyChosenBits[bit] = np.array([], dtype=np.int64)


    def calculateParts(self, cell):
        positivePart = 0
        negativePart = 0
        if np.random.rand() < self.Pp[cell]:
            positivePart = self.ac.DELTA_I
        if np.random.rand() < self.Pn[cell]:
            negativePart = -self.ac.DELTA_I
        if positivePart > 0:
            self.Pp[cell] -= self.ac.P_P_CHANGE
            self.Pn[cell] += self.ac.P_N_CHANGE
        if negativePart < 0:
            self.Pn[cell] -= self.ac.P_N_CHANGE
            self.Pp[cell] += self.ac.P_P_CHANGE

        return positivePart, negativePart

    def digitizeValue(self, value):
        i = np.digitize(value, self.centers, right=True)
        return self.bins[i]

    def movingBitNextValue(self, t, a, b, c):
        noise = np.random.normal(0, self.ac.MOVING_BIT_NOISE_STD, 1)[0]  # Gaussian white noise
        analogValue = a * np.exp(b * t) + c + noise
        return self.digitizeValue(analogValue)

    def erraticBitNextValue(self, bit):
        if bit in self.activeErraticBits:
            if(self.activeErraticBitsTime[bit] < 10):
                self.activeErraticBitsTime[bit] += 1
                return self.values[bit]
            else:
                self.activeErraticBitsTime[bit] = 0
                self.activeErraticBits = np.setdiff1d(self.activeErraticBits, np.array([bit]))
                return self.historyChosenBits[bit][len(self.historyChosenBits[bit]) - 12]

        if np.random.choice(2,1,p=[self.ac.ERRATIC_BIT_CHANCE, 1 - self.ac.ERRATIC_BIT_CHANCE]) == 0:
            self.activeErraticBits = np.append(self.activeErraticBits, bit)
            self.activeErraticBitsTime[bit] = 0
            return np.random.normal(8, 1, 1)[0]
        return self.values[bit]

    def progressTimeOverFewCells(self):
        bitsToFlip = np.random.choice(self.ac.FLASH_SIZE_KB * 1024 * 8, self.ac.NMB_BITS_TO_PROCESS)
        for movingBit in self.movingBits:
            self.values[movingBit] = self.movingBitNextValue(self.movingBitTime[movingBit], self.initialValues[movingBit],
                                                             self.interpolationB[movingBit], self.interpolationC[movingBit])
        for erraticBit in self.erraticBits:
            self.values[erraticBit] = self.erraticBitNextValue(erraticBit)

        for historyBit in self.ac.HISTORY_BITS:
            self.historyChosenBits[historyBit] = np.append(self.historyChosenBits[historyBit], self.values[historyBit])
            np.save("data", self.historyChosenBits)

        bitsToFlip = np.append(bitsToFlip, 82)
        for cellIndex in bitsToFlip:
            if self.values[cellIndex] == 0.5:
                continue

            if cellIndex in self.movingBits:
                continue

            pP, nP = self.calculateParts(cellIndex)
            self.values[cellIndex] = self.digitizeValue(self.values[cellIndex] + pP + nP)

    def addHistoryBit(self, bit):
        self.ac.HISTORY_BITS = np.append(self.ac.HISTORY_BITS, bit)
        self.historyChosenBits[bit] = np.array([])
        # Populate with old "madeup" data to make time correct
        for value in self.historyChosenBits[self.ac.HISTORY_BITS[0]]:
            self.historyChosenBits[bit] = np.append(self.historyChosenBits[bit], self.values[bit])

    def progressTime(self):
        if self.ac.REPEAT_CSV_FILES:
            self.t += 1
            return
        self.progressTimeOverFewCells()
        self.t += 1
        for bit in self.movingBits:
            self.movingBitTime[bit] += 1

    def mean(self):
        s = 0
        c = 0
        for cellIndex, cell in np.ndenumerate(self.values):
            if cell != 0.5:
                s += cell
                c += 1
        return s / c

    def sweep(self, threshold):
        r = []
        for index, value in np.ndenumerate(self.values):
            if value == 0.5:
                continue
            if value < threshold:
                r.append(index[0])
        return r

    def program(self, bit):
        self.values[bit] = np.random.normal(self.ac.ICELL_MEAN, self.ac.ICELL_STD, 1)[0]
        self.Pp[bit] = self.ac.P_PN_INITIAL
        self.Pn[bit] = self.ac.P_PN_INITIAL
        self.initialValues[bit] = self.values[bit]
        if bit in self.movingBits:
            self.interpolationB[bit] = np.random.normal(self.ac.INTERPOLATION_B, self.ac.INTERPOLATION_B_STD, 1)[0]
            self.interpolationC[bit] = np.random.normal(self.ac.INTERPOLATION_C, self.ac.INTERPOLATION_C_STD, 1)[0]
            self.initialValues[bit] -= self.interpolationC[bit]
            self.movingBitTime[bit] = 0

        print("Reprogrammed {} to {}".format(bit, self.values[bit]))

class SimulationManager(object):
    def __init__(self):
        self.activeSimulations = {}

    @staticmethod
    def generateId():
        return uuid.uuid4().int & (1 << 8) - 1

    def requestNewSimulation(self):
        sim_nmb = self.generateId()
        while sim_nmb in self.activeSimulations.keys():
            # TODO: Add check for if the total number of allowed uuids has been exceeded
            sim_nmb = self.generateId()

        self.activeSimulations[sim_nmb] = Simulation()

        return sim_nmb
